-- table utility functions --

local type = type
local pairs = pairs

local substr = string.sub
local str_replace = string.gsub

local tbl_count = table_size
local arr_insert = table.insert
local arr_sort = table.sort
local arr_unpack = table.unpack
local arr_empty = {}

local function is_not_table(v)
	local t = type(v)
	return t ~= "table" and (t ~= "userdata" or v.object_name ~= "LuaCustomTable")
end

local function tbl_copy(...)
	local lookup = {}
	local function tcp(tbl)
		if is_not_table(tbl) then return tbl end
		if lookup[tbl] then return lookup[tbl] end
		local ret = {}
		lookup[tbl] = ret
		for k,v in pairs(tbl) do
			ret[tcp(k)] = tcp(v)
		end
		return ret
	end
	return tcp(...)
end

local function tbl_eq(...)
	local lookup = {}
	local function teq(a,b)
		if a == b then return true end -- short circuit on the same table/value
		if is_not_table(a) or is_not_table(b) then return false end -- only if both are tables they can be equal
		if lookup[a] and lookup[a][b] or lookup[b] and lookup[b][a] then return true end -- recursive tables already under investigation are equal until proven otherwise
		if lookup[a] then -- keep lookup small
			lookup[a][b] = true
		elseif lookup[b] then -- keep lookup small
			lookup[b][a] = true
		else
			lookup[a] = {[b]=true}
		end
		for k,v in pairs(a) do
			if not teq(b[k],v) then
				return false
			end
		end
		for k,v in pairs(b) do
			if a[k] == nil and v ~= nil then
				return false
			end
		end
		return true
	end
	return teq(...)
end

local function arr_add_new(arr,val)
	for _,v in pairs(arr) do
		if tbl_eq(v,val) then return v end
	end
	arr_insert(arr,val)
	return val
end

local function tbl_values(tbl,insert_func,sort_func)
	if insert_func == nil then insert_func = arr_insert end
	local arr = {}
	for _,v in pairs(tbl) do
		insert_func(arr,v)
	end
	if sort_func then
		arr_sort(arr,sort_func)
	end
	return arr
end





-- self accelerating functions --

local function get_player(...)
	get_player = game.get_player
	return get_player(...)
end

local function get_logistic_sections(ntt)
	local logi = ntt.get_logistic_sections()
	return logi and logi.sections or arr_empty
end
local function has_sortable_logistic_sections(...)
	local forbidden_section_type = {
		[defines.logistic_section_type.transitional_request_controlled] = true,
	}
	has_sortable_logistic_sections = function(ntt)
		local logi = ntt.get_logistic_sections()
		if not logi then return false end
		logi = logi.sections
		if not logi then return false end
		for _,sect in pairs(logi) do
			if forbidden_section_type[sect.type] then return false end
		end
		return true
	end
	return has_sortable_logistic_sections(...)
end





-- compare functions --
local function cmp_str(a,b)
	local x = a or ""
	local y = b or ""
	if x > y then return false end
	if x < y then return true end
	return nil
end
local function cmp_comparator(...)
	local ComparatorOrder = {}
	do
		for k,v in pairs{{"<"},{"<=","≤"},{">=","≥"},{">"},{"!=","≠"},{"="}} do
			for _,c in pairs(v) do
				ComparatorOrder[c] = k
			end
		end
	end
	cmp_comparator = function(a,b)
		local x = ComparatorOrder[a] or 0
		local y = ComparatorOrder[b] or 0
		if x > y then return false end
		if x < y then return true end
		return nil
	end
	return cmp_comparator(...)
end
local function cmp_proto(x,y)
	if x.group.order > y.group.order then return false end
	if x.group.order < y.group.order then return true end
	if x.group.name > y.group.name then return false end
	if x.group.name < y.group.name then return true end
	if x.subgroup.order > y.subgroup.order then return false end
	if x.subgroup.order < y.subgroup.order then return true end
	if x.subgroup.name > y.subgroup.name then return false end
	if x.subgroup.name < y.subgroup.name then return true end
	if x.order > y.order then return false end
	if x.order < y.order then return true end
	if x.name > y.name then return false end
	if x.name < y.name then return true end
	if x.type > y.type then return false end
	if x.type < y.type then return true end
	return nil
end
local function cmp_qual(x,y)
	if x.level > y.level then return false end
	if x.level < y.level then return true end
	return cmp_proto(x,y)
end
local function cmp_prototype_str(...)
	local lookup = {}
	do -- nothing in here shall generate upvalues
		local protos = {}
		local order = {}
		for n,p in pairs{
			["asteroid-chunk"]="asteroid_chunk",
			["entity"]="entity",
			["equipment"]="equipment",
			["fluid"]="fluid",
			["item"]="item",
			["quality"]="quality",
			["recipe"]="recipe",
			["space-connection"]="space_connection",
			["space-location"]="space_location",
			["technology"]="technology",
			["tile"]="tile",
			["virtual"]="virtual_signal",
		} do
			lookup[n] = {}
			for k,v in pairs(prototypes[p]) do
				arr_insert(order,v)
				protos[v] = {n,k}
			end
		end
		arr_sort(order,cmp_proto)
		for k,v in pairs(order) do
			local t,n = arr_unpack(protos[v])
			lookup[t][n] = k
		end
	end -- nothing in here shall generate upvalues
	cmp_prototype_str = function(aType,aName,bType,bName)
		if not (aType and aName) then
			return bType and bName and true or nil
		end
		if not (bType and bName) then
			return false
		end
		if aType==bType and aName==bName then
			return nil
		end
		local x,y = lookup[aType][aName],lookup[bType][bName]
		if x then
			if y then
				return x < y
			end
			return false
		end
		return y and true or nil
	end
	return cmp_prototype_str(...)
end
local function cmp_quality_str(...)
	local lookup = {}
	do -- nothing in here shall generate upvalues
		local order = {}
		for _,v in pairs(prototypes["quality"]) do
			arr_insert(order,v)
		end
		arr_sort(order,cmp_qual)
		for k,v in pairs(order) do
			lookup[v.name] = k
		end
	end -- nothing in here shall generate upvalues
	cmp_quality_str = function(a,b)
		local x,y = lookup[a],lookup[b]
		if x then
			if y then
				return x < y
			end
			return false
		end
		return y and true or nil
	end
	return cmp_quality_str(...)
end

-- sort functions --

local function sort_logistics(a,b)
	local x = a.value
	local y = b.value
	local c = cmp_prototype_str( x.type,x.name , y.type,y.name )
	if c ~= nil then return c end
	c = cmp_comparator(x.comparator,y.comparator)
	if c ~= nil then return c end
	return cmp_quality_str(x.quality,y.quality)
end
local function sort_inf_filter(a,b)
	local c = cmp_prototype_str( "item",a.name , "item",b.name )
	if c ~= nil then return c end
	c = cmp_str(a.mode,b.mode)
	if c ~= nil then return c end
	return cmp_quality_str(a.quality,b.quality)
end
local function sort_decon_filter(a,b)
	local c = cmp_prototype_str( "entity",a.name , "entity",b.name )
	if c ~= nil then return c end
	c = cmp_comparator(a.comparator,b.comparator)
	if c ~= nil then return c end
	return cmp_quality_str(a.quality,b.quality)
end
local function sort_decon_tile_filter(a,b)
	return cmp_prototype_str( "tile",a , "tile",b )
end
local function sort_upgrade_filter(a,b)
	local c = cmp_prototype_str( a.type,a.name , b.type,b.name )
	if c ~= nil then return c end
	c = cmp_comparator(a.comparator,b.comparator)
	if c ~= nil then return c end
	return cmp_quality_str(a.quality,b.quality)
end



-- configuration helper --

local anchors = {}
do
	local pos = defines.relative_gui_position.right
	for k,v in pairs(defines.relative_gui_type) do
		anchors[str_replace(substr(k,1,-5),"_","-")] = {gui=v,position=pos}
	end
	anchors["blueprint"] = anchors["blueprint-setup"]
	anchors["cargo-wagon"] = anchors["container"]
	anchors["deconstruction-planner"] = anchors["deconstruction-item"]
	anchors["infinity-cargo-wagon"] = anchors["container"]
	anchors["infinity-container"] = anchors["container"]
	anchors["logistic-container"] = anchors["container"]
	anchors["upgrade-planner"] = anchors["upgrade-item"]
end
local special_anchors = {
	["player"]="other-player",
	["self"]="controller",
}
local has_infinity_filter = {
	["infinity-cargo-wagon"]=true,
	["infinity-container"]=true,
}
local skip_logistic_filter = {
	["rocket"]=true,
}
local inv_config = {
	["asteroid-collector"]={defines.inventory.chest},
	["car"]={defines.inventory.car_trunk,defines.inventory.car_trash},
	["cargo-landing-pad"]={defines.inventory.cargo_landing_pad_main,defines.inventory.cargo_landing_pad_trash},
	["cargo-wagon"]={defines.inventory.cargo_wagon},
	["character"]={defines.inventory.character_main,defines.inventory.character_trash},
	["self"]={defines.inventory.character_main,defines.inventory.character_trash},
	["container"]={defines.inventory.chest},
	["infinity-cargo-wagon"]={defines.inventory.cargo_wagon},
	["infinity-container"]={defines.inventory.chest},
	["linked-container"]={defines.inventory.chest},
	["logistic-container"]={defines.inventory.chest,defines.inventory.logistic_container_trash},
	["proxy-container"]={defines.inventory.proxy_main},
	["roboport"]={defines.inventory.roboport_robot,defines.inventory.roboport_material},
	["rocket-silo"]={defines.inventory.rocket_silo_rocket,defines.inventory.rocket_silo_trash},
	["space-platform-hub"]={defines.inventory.hub_main,defines.inventory.hub_trash},
	["spider-vehicle"]={defines.inventory.spider_trunk,defines.inventory.spider_trash},
}
local evt_gui_open_close_item_map = {
	["asteroid-collector"]="chest",
	["blueprint-book"]="book",
	["car"]="car",
	["cargo-landing-pad"]="space",
	["cargo-wagon"]="train",
	["character"]="player",
	["constant-combinator"]="math",
	["container"]="chest",
	["deconstruction-planner"]="decon",
	["infinity-cargo-wagon"]="train",
	["infinity-container"]="chest",
	["linked-container"]="chest",
	["logistic-container"]="chest",
	["proxy-container"]="chest",
	["roboport"]="robo",
	["rocket-silo"]="rocket",
	["self"]="self",
	["space-platform-hub"]="space",
	["spider-vehicle"]="car",
	["upgrade-planner"]="upgrade",
}

local sort_data = {}





local sort_planner_functions

-- sort inventories --

local function handle_entity_inventory(plr,ntt,num)
	local inv = inv_config[ntt.type]
	if not inv then return end
	inv = inv[num]
	if not inv then return end
	inv = ntt.get_inventory(inv)
	if not inv then return end
	inv.sort_and_merge()
end

-- sort logistic section filters --

local function handle_entity_logistic_sections(plr,ntt)
	for _,sect in pairs(get_logistic_sections(ntt)) do
		if sect.is_manual and sect.filters and sect.filters_count > 1 then
			local lst = {}
			for _,fltr in pairs(sect.filters) do
				if fltr.value then
					arr_add_new(lst,fltr)
				end
			end
			arr_sort(lst,sort_logistics)
			sect.filters = lst
		end
	end
end

-- sort infinity container filters --

local function handle_entity_infinity_filters(plr,ntt)
	if ntt.infinity_container_filters and tbl_count(ntt.infinity_container_filters) > 1 then
		local lst = {}
		for _,fltr in pairs(ntt.infinity_container_filters) do
			if fltr.name then
				arr_add_new(lst,fltr)
			end
		end
		arr_sort(lst,sort_inf_filter)
		for i,fltr in pairs(lst) do
			fltr.index = i
		end
		ntt.infinity_container_filters = lst
	end
end

-- sort deconstruction planner --

local function handle_decon_planner(plr,itm)
	local lst = {}
	for n,fltr in pairs(itm.entity_filters) do
		arr_add_new(lst,fltr)
	end
	arr_sort(lst,sort_decon_filter)
 	itm.entity_filters = lst
	-- also sort tiles
	lst = {}
	for _,fltr in pairs(itm.tile_filters) do
		arr_add_new(lst,fltr)
	end
	arr_sort(lst,sort_decon_tile_filter)
	itm.tile_filters = lst
end

-- sort upgrade planner --

local function handle_upgrade_planner(plr,itm)
	local from = {}
	local to = {}
	for i = 1,itm.mapper_count do
		local s = itm.get_mapper(i,"from")
		local d = itm.get_mapper(i,"to")
		if s and s.name and d and d.name then
			to[arr_add_new(from,s)] = d
		end
	end
	arr_sort(from,sort_upgrade_filter)
	for i = 1,itm.mapper_count do
		itm.set_mapper(i,"from",from[i])
		itm.set_mapper(i,"to",to[from[i]])
	end
end

-- sort blueprint book --

local function handle_blueprint_book(plr,itm)
	if itm.object_name ~= "LuaRecord" then return end
	local cond = {}
	for _,v in pairs{"decon","upgrade",--[["book",]]"books"} do
		cond[v] = plr.mod_settings["sbSortedLogistics-sort-"..v].value
	end
	local sort_fn = tbl_copy(sort_planner_functions)
	sort_fn["book"] = function(plr,book)
		-- sort contents when finally available
		--if cond["book"] then
		--end
		if not cond["books"] then return end
		for _,rec in pairs(book.contents) do
			if rec and cond[evt_gui_open_close_item_map[rec.type]] then
				local f = sort_fn[evt_gui_open_close_item_map[rec.type]]
				if f then f(plr,rec) end
			end
		end
	end
	return sort_fn["book"](plr,itm)
end

-- sort planner easy access --

sort_planner_functions = {
	["upgrade"]=handle_upgrade_planner,
	["decon"]=handle_decon_planner,
	["book"]=handle_blueprint_book,
}










-- gui functions --

local function gui_destroy(plr)
	local plr_gui = plr.gui.relative["sbSortedLogistics_gui"]
	if plr_gui then
		plr_gui.destroy()
	end
end
local function gui_show(plr,anchor,elements)
	gui_destroy(plr)
	if #elements<1 then return end
	local plr_gui = plr.gui.relative
	local frame = plr_gui.add{
		type="frame",
		name="sbSortedLogistics_gui",
		caption={"sbSortedLogistics.gui_frame"},
		direction="vertical",
		anchor=anchor,
	}
	for k,v in pairs(elements) do
		frame.add(v)
	end
end
local function gui_show_entity(plr,ntt,typ,when)
	sort_data[plr.index] = {ntt,typ}
	local elements = {}

	for k,v in pairs(inv_config[ntt.type] or arr_empty) do
		local inv = ntt.get_inventory(v)
		if inv and #inv>1 and (when == "always" or not plr.mod_settings["sbSortedLogistics-sort-"..typ..k].value) then
			arr_insert(elements,{
				type="button",
				name="sbSortedLogistics_sort"..k,
				caption={"sbSortedLogistics.btn_sort"},
				tooltip={"sbSortedLogistics.desc-"..typ..k},
			})
			if when ~= "mini" then
				arr_insert(elements,{
					type="checkbox",
					state=plr.mod_settings["sbSortedLogistics-sort-"..typ..k].value,
					name="sbSortedLogistics_auto_"..typ..k,
					caption={"mod-setting-name.sbSortedLogistics-sort-"..typ..k},
					tooltip={"mod-setting-description.sbSortedLogistics-sort-"..typ..k},
				})
			end
		end
	end
	if has_sortable_logistic_sections(ntt) and (when == "always" or not plr.mod_settings["sbSortedLogistics-sort-"..typ].value) then
		arr_insert(elements,{
			type="button",
			name="sbSortedLogistics_sortl",
			caption={"sbSortedLogistics.btn_sort"},
			tooltip={"sbSortedLogistics.desc-"..typ},
		})
		if when ~= "mini" then
			arr_insert(elements,{
				type="checkbox",
				state=plr.mod_settings["sbSortedLogistics-sort-"..typ].value,
				name="sbSortedLogistics_auto_"..typ,
				caption={"mod-setting-name.sbSortedLogistics-sort-"..typ},
				tooltip={"mod-setting-description.sbSortedLogistics-sort-"..typ},
			})
		end
	end
	if has_infinity_filter[ntt.type] and (when == "always" or not plr.mod_settings["sbSortedLogistics-sort-infinity"].value) then
		arr_insert(elements,{
			type="button",
			name="sbSortedLogistics_sorti",
			caption={"sbSortedLogistics.btn_sort"},
			tooltip={"sbSortedLogistics.desc-infinity"}
		})
		if when ~= "mini" then
			arr_insert(elements,{
				type="checkbox",
				state=plr.mod_settings["sbSortedLogistics-sort-infinity"].value,
				name="sbSortedLogistics_auto_infinity",
				caption={"mod-setting-name.sbSortedLogistics-sort-infinity"},
				tooltip={"mod-setting-description.sbSortedLogistics-sort-infinity"},
			})
		end
	end

	return gui_show(plr,anchors[special_anchors[typ] or ntt.type],elements)
end
local function gui_show_planner(plr,bp,typ,when)
	sort_data[plr.index] = {bp,typ}
	local elements = {}

	if when == "always" or not plr.mod_settings["sbSortedLogistics-sort-"..typ].value then
		if typ ~= "book" then
			arr_insert(elements,{
				type="button",
				name="sbSortedLogistics_sortp",
				caption={"sbSortedLogistics.btn_sort"},
				tooltip={"sbSortedLogistics.desc-"..typ},
			})
		end
		if when ~= "mini" then
			for _,t in pairs(typ=="book" and {"decon","upgrade",--[[typ,]]"books"} or {typ}) do
				arr_insert(elements,{
					type="checkbox",
					state=plr.mod_settings["sbSortedLogistics-sort-"..t].value,
					name="sbSortedLogistics_auto_"..t,
					caption={"mod-setting-name.sbSortedLogistics-sort-"..t},
					tooltip={"mod-setting-description.sbSortedLogistics-sort-"..t},
				})
			end
		end
	end

	return gui_show(plr,anchors[bp.type],elements)
end
local gui_click = {
	["sbSortedLogistics_sort1"] = function(el,plr,obj,typ)
		return handle_entity_inventory(plr,obj,1)
	end,
	["sbSortedLogistics_sort2"] = function(el,plr,obj,typ)
		return handle_entity_inventory(plr,obj,2)
	end,
	["sbSortedLogistics_sortl"] = function(el,plr,obj,typ)
		return handle_entity_logistic_sections(plr,obj)
	end,
	["sbSortedLogistics_sorti"] = function(el,plr,obj,typ)
		return handle_entity_infinity_filters(plr,obj)
	end,
	["sbSortedLogistics_sortp"] = function(el,plr,obj,typ)
		return sort_planner_functions[typ](plr,obj)
	end,
}
do
	local function mk_gui_click_auto(id)
		local setting_name = "sbSortedLogistics-sort-"..id
		gui_click["sbSortedLogistics_auto_"..id] = function(el,plr)
			plr.mod_settings[setting_name] = {value=el.state}
		end
	end
	for k,v in pairs{"infinity","self","player"} do
		mk_gui_click_auto(v)
	end
	for k,v in pairs(special_anchors) do
		mk_gui_click_auto(k)
	end
	for k,v in pairs(tbl_values(evt_gui_open_close_item_map,arr_add_new)) do
		mk_gui_click_auto(v)
	end
	for k,v in pairs(inv_config) do
		for i,j in pairs(v) do
			local id = evt_gui_open_close_item_map[k]..i
			mk_gui_click_auto(evt_gui_open_close_item_map[k]..i)
		end
	end
end





-- handle gui open and auto sort

local function evt_handle_entity(plr,open,obj,typ)
	for k,v in pairs(inv_config[obj.type] or arr_empty) do
		if plr.mod_settings["sbSortedLogistics-sort-"..typ..k].value then
			handle_entity_inventory(plr,obj,k)
		end
	end
	if not skip_logistic_filter[typ] and plr.mod_settings["sbSortedLogistics-sort-"..typ].value then
		handle_entity_logistic_sections(plr,obj)
	end
	if has_infinity_filter[obj.type] and plr.mod_settings["sbSortedLogistics-sort-infinity"].value then
		handle_entity_infinity_filters(plr,obj)
	end
	local when = plr.mod_settings["sbSortedLogistics-gui-show"].value
	if open and when ~= "never" then
		return gui_show_entity(plr,obj,typ,when)
	end
end
local function evt_handle_planner(plr,open,obj,typ)
	if plr.mod_settings["sbSortedLogistics-sort-"..((typ=="book") and "books" or typ)].value then
		sort_planner_functions[typ](plr,obj)
	end
-- 	if plr.mod_settings["sbSortedLogistics-sort-"..typ].value or (typ == "book" and plr.mod_settings["sbSortedLogistics-sort-books"].value) then
-- 		sort_planner_functions[typ](plr,obj)
-- 	end
	local when = plr.mod_settings["sbSortedLogistics-gui-show"].value
	if open and when ~= "never" then
		return gui_show_planner(plr,obj,typ,when)
	end
end





-- branch out gui open close event --

local controller_character = defines.controllers.character
local evt_gui_open_close_func = {
	[defines.gui_type.controller]=function(ev,plr,open)
		local chr = plr.character
		if chr and plr.controller_type == controller_character then
			return evt_handle_entity(plr,open,chr,"self")
		end
	end,
	[defines.gui_type.entity]=function(ev,plr,open)
		local ntt = ev.entity
		local typ = evt_gui_open_close_item_map[ntt.type]
		if typ then
			return evt_handle_entity(plr,open,ntt,typ)
		end
	end,
	[defines.gui_type.item]=function(ev,plr,open)
		local bp = ev.item
		local typ = evt_gui_open_close_item_map[bp.name]
		if typ then
			return evt_handle_planner(plr,open,bp,typ)
		end
	end,
}



script.on_event(defines.events.on_gui_click, function(pEvent)
	local fn = gui_click[pEvent.element.name]
	local dat = sort_data[pEvent.player_index]
	if fn and dat then
		return fn(pEvent.element,get_player(pEvent.player_index),arr_unpack(dat))
	end
end)
script.on_event(defines.events.on_gui_opened, function(pEvent)
	local plr = get_player(pEvent.player_index)
	gui_destroy(plr)
	sort_data[plr.index] = nil
	local fn = evt_gui_open_close_func[pEvent.gui_type]
	if fn then return fn(pEvent,plr,true) end
end)
script.on_event(defines.events.on_gui_closed, function(pEvent)
	local plr = get_player(pEvent.player_index)
	gui_destroy(plr)
	sort_data[plr.index] = nil
	local fn = evt_gui_open_close_func[pEvent.gui_type]
	if fn then return fn(pEvent,plr,false) end
end)
script.on_event(defines.events.on_player_cursor_stack_changed, function(pEvent)
	local plr = get_player(pEvent.player_index)
	local fn
	local itm = plr.cursor_stack
	if itm and itm.valid_for_read then
		fn = sort_planner_functions[evt_gui_open_close_item_map[itm.name]]
	else
		itm = plr.cursor_record
		if itm and itm.valid_for_write then
			fn = sort_planner_functions[evt_gui_open_close_item_map[itm.type]]
		end
	end
	if fn then return fn(plr,itm) end
end)
